/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include <stdexcept>
#include "file.hpp"

using namespace std;

CFile::CFile(const char* filename,const char* mode):
  file(NULL)
{
  file = fopen(filename,mode);
  if (file == NULL)
    throw runtime_error("Unable to open file");
}

void CFile::Close()
{
  if (file)
    fclose(file);
  file = NULL;
}

void CFile::Seek(long offset,int whence) const
{
  if ( fseek(file,offset,whence)!=0 )
    throw runtime_error("Unable to seek");
}

long CFile::Tell() const
{
  long pos = ftell(file);
  if ( pos ==  -1L )
    throw runtime_error("Seek error");
  return pos;
}

char *CFile::ReadString(int delim) const
{
  char *pbegin = new char[16];
  char *p      = pbegin;
  char *pend   = pbegin + 16;
  int c;
  for (;;)
  {
    c = getc(file);
    if (c==EOF)
    {
      delete [] pbegin;
      throw runtime_error("Error reading string");
    }
    if (p==pend) {
      int size = (pend-pbegin);
      pend = new char[size+16];
      memcpy(pend,pbegin,size*sizeof(char));
      delete [] pbegin;
      pbegin = pend;
      pend = (p = pbegin + size) + 16;
    }
    *p++ = (char)c;
    if (c==delim) {
      p[-1] = '\0';
      return pbegin;
    }
  }
}

// !!! problem with portability ?? See charBit and uint16 conversion
char *CFile::ReadStringFlag(unsigned flag) const
{
  unsigned char charBit = (flag & strUcs2) ? 1:0;
  char *pbegin = new char[16<<charBit];
  char *p      = pbegin;
  char *pend   = pbegin + (16<<charBit);
  int c,cc;
  unsigned len = ~0u;

  uint32_t ui32;
  uint16_t ui16;
  uint8_t  ui8;

  switch (flag & strLenMask)
  {
  case strLen8:  RawRead(&ui8, sizeof(ui8) ); len = ui8;  break;
  case strLen16: RawRead(&ui16,sizeof(ui16)); len = ui16; break;
  case strLen32: RawRead(&ui32,sizeof(ui32)); len = ui32; break;
  }

  cc = 0;
  for (;;--len)
  {
    // extend buffer if needed
    if (p==pend) {
      int size = (pend-pbegin);
      pend = new char[size+(16<<charBit)];
      memcpy(pend,pbegin,size*sizeof(char));
      delete [] pbegin;
      pbegin = pend;
      pend = (p = pbegin + size) + (16<<charBit);
    }

    // check end string
    if (len == 0)
    {
      *p++ = 0;
      if (charBit)
        *p++ = 0;
      break;
    }

    // read char
    c = getc(file);
    if (c != EOF && charBit)
    {
      cc = fgetc(file);
      if (cc==EOF)
        c = EOF;
    }
    if (c==EOF)
    {
      delete [] pbegin;
      throw runtime_error("Error reading string");
    }

    // write char
    *p++ = (char)c;
    if (charBit)
      *p++ = (char)cc;

    if ( (flag&strLenMask) == 0 )
    {
      if ( c==0 && cc==0 )
        break;
    }
  }

  return pbegin;
}

void CFile::RawRead(void *ptr,size_t size) const
{
  if ( fread(ptr,1,size,file) != size )
    throw runtime_error("Unable to read from file");
}

void CFile::RawWrite(const void *ptr,size_t size)
{
  if ( fwrite(ptr,1,size,file) != size )
    throw runtime_error("Unable to write to file");
}

CFile::~CFile()
{
  Close();
}
