/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#ifndef FILE_PEFILE_HPP
#define FILE_PEFILE_HPP

#include "peformat.h"
#include "file.hpp"

typedef uint16_t unicode_t;

class RVAFileTranslator;

typedef struct PeResourceId
{
  unicode_t* name; // name of directory
  unsigned id;   // valid if name == 0
} PeResourceId;

typedef struct PeResourceDirectoryInfo: public PeResourceId
{
  PE_IMAGE_RESOURCE_DIRECTORY dir;
  PeResourceDirectoryInfo* highLevel;
} PeResourceDirectoryInfo;

typedef struct PeResourceDirectoryData: public PeResourceId
{
  PE_IMAGE_RESOURCE_DATA_ENTRY data;
} PeResourceDirectoryData;

class PeFile
{
public:
  PeFile(ICFile& file);
  ~PeFile();

  // headers access
  const PE_IMAGE_FILE_HEADER&     GetFileHeader() const
  { return nt_hdr.FileHeader; };
  const PE_IMAGE_OPTIONAL_HEADER& GetOptionalHeader() const
  { return nt_hdr.OptionalHeader; };

  // section access
  const PE_IMAGE_SECTION_HEADER*  GetSections() const
  { return sections; };
  const PE_IMAGE_SECTION_HEADER&  GetSection(int n) const
  { return sections[n]; };
  unsigned GetSectionCount() const
  { return numSection; };

  // directory access
  const PE_IMAGE_DATA_DIRECTORY& GetDataDirectory(unsigned n) const
  { return GetOptionalHeader().DataDirectory[n]; };

  // other access
  uint32_t GetImageBase() const
  { return GetOptionalHeader().ImageBase; };

  // write to stdout
  static void WriteFileHdr(const PE_IMAGE_FILE_HEADER& hdr);
  static void WriteOptionalHdr(const PE_IMAGE_OPTIONAL_HEADER& hdr,
                               const RVAFileTranslator& rva);
  static void WriteSectionHdr(const PE_IMAGE_SECTION_HEADER& hdr);
  static void WriteExport(uint32_t address,uint32_t hint,const char* dll_name,const char* func_name,void* param);
  static void WriteImport(uint32_t address,uint32_t hint,const char* dll_name,const char* func_name,void* param);
  static void WriteRelocation(uint32_t address,unsigned type,void* param);
  void WriteResource();

  // directory parsing
  typedef void (*PImportProc)(uint32_t address,uint32_t hint,const char* dll_name,const char* func_name,void* param);
  void ParseImport(PImportProc proc,void* param);
  typedef void (*PRelocationProc)(uint32_t address,unsigned type,void* param);
  void ParseRelocations(PRelocationProc proc,void* param);

  // called by ParseResource for every resource
  //   file is the pe file. File position point to resource data if data available
  //   data pointer to data information. If directory data == NULL
  //   info point to directory information. Valid also for data information
  //   level level of directory. First time called with level == 0 and data == NULL (init)
  //   param user parameter
  typedef void (*PResourceProc)(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param);
  void ParseResource(PResourceProc proc,void* param);

  typedef void (*PExportProc)(uint32_t address,uint32_t hint,const char* dll_name,const char* func_name,void* param);
  void ParseExport(PExportProc proc,void* param);

//  RVAFileTranslator& GetRVA() { return *(RVAFileTranslator*)rva; }
  RVAFileTranslator& GetRVA() { return *rva; }
//  const RVAFileTranslator& GetRVA() const { return *(const RVAFileTranslator*)rva; }
  const RVAFileTranslator& GetRVA() const { return *rva; }
private:
  PE_IMAGE_NT_HEADERS nt_hdr;
  PE_IMAGE_SECTION_HEADER *sections;
  unsigned numSection;
  ICFile& file;
//  char rva[sizeof(RVAFileTranslator)];
  RVAFileTranslator* rva;
};
#endif //FILE_PEFILE_HPP
