/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include "persrc.h"
#include <stdexcept>
#include <list>
#include "utils/f77auto_ptr"

extern int out_ws(const unicode_t* s);

struct iconinfo
{
  // id of icon must be numeric
  unsigned id;
  long pos;
  uint32_t size;
};

typedef std::list<iconinfo> TIconInfos;

static void CollectIconInfos(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  _DEBUG_(fprintf(stderr,"data %p level %d name %p id %d\n", data, level, info->name, info->id ));
  if (data != NULL && level == 3 && info->name == NULL && info->highLevel->id == 3)
  {
    _DEBUG_(fprintf(stderr,"---data %p level %d name %p id %d\n", data, level, info->name, info->highLevel->id));
    iconinfo iinfo;
    iinfo.id = info->id;
    iinfo.pos = file.Tell();
    iinfo.size = data->data.Size;
    _DEBUG_(fprintf(stderr,"---id %d pos %ld size %d\n", iinfo.id, iinfo.pos, iinfo.size));
    
    (*reinterpret_cast<TIconInfos*>(param)).push_back(iinfo);
  }
}

struct WriteIconParam
{
  TIconInfos iconInfos;
  int n;
  WriteIconParam():n(1) {};

};
  
static const iconinfo* findIconInfo(const TIconInfos& infos,unsigned id)
{
  TIconInfos::const_iterator end = infos.end();
  for(TIconInfos::const_iterator i = infos.begin(); i != end; ++i)
    if ( (*i).id == id )
      return &(*i);
  return NULL;
}

static void CopyFileToFile(const ICFile& from,ICFile& to,size_t bytes)
{
  char buffer[128];
  while(bytes > 0)
  {
    size_t len = f77_min(bytes,(size_t)128);
    from.RawRead(buffer,len);
    to.RawWrite(buffer,len);
    bytes -= len;
  }
}

static void WriteIcons(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  static const char zeroes[2] = { 0, 0 };
  if (data != NULL && level == 3 && info->highLevel->id == 14)
  {
    WriteIconParam *p = reinterpret_cast<WriteIconParam*>(param);
    
    // ok, write this icon
    NEWHEADER hdr;
    RawRead(file,hdr);
    
    char buffer[128];
    sprintf(buffer,"resico%d.ico",p->n);
    ++p->n;
    CFile ico(buffer,"wb");
    
    // save header
    RawWrite(ico,hdr);
    
    
    freddy77::auto_delete_ptr<const iconinfo*> icons(new const iconinfo*[hdr.ResCount]);
    int i;
    for( i = 0; i < hdr.ResCount; ++i )
    {
      RESDIR resico;
      RawRead(file,resico);
      
      // save icon header
      RawWrite(ico,resico);
      ico.RawWrite(zeroes,2);
      
      // find icon and save it
      const iconinfo* info = findIconInfo(p->iconInfos,resico.IconCursorId);
      *(icons.get()+i) = info;
      if (!info)
        throw std::runtime_error("icon not found for icon group");
      if (resico.BytesInRes != info->size)
        throw std::runtime_error("icon with wrong size");
      // !!! check icon dimension and colors
    }
    for( i = 0; i < hdr.ResCount; ++i )
    {
      const iconinfo* info = *(icons.get()+i);
      file.Seek(info->pos);
      CopyFileToFile(file,ico,info->size);
    }
  }
}

void ResourceExtract::ExtractIcons(PeFile& peFile)
{
//  fprintf(stderr,"sizeof hdr %d\n",sizeof(RESDIR));
//  throw std::runtime_error("ferma");
  // parse first for information retrieval (position and size of resources)
  WriteIconParam param;
  peFile.ParseResource(CollectIconInfos,&param.iconInfos);

  // parse for output files
  peFile.ParseResource(WriteIcons,&param);
}

struct WriteCurBMPIconParam
{
  unsigned int type, n;
  WriteCurBMPIconParam(int t) : type(t), n(0) {}
};

static void WriteCurBMPIconInfo(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  WriteCurBMPIconParam *p = reinterpret_cast<WriteCurBMPIconParam*>(param);
  if (data != NULL && level == 3 && info->highLevel->id == p->type)
  {
    if (p->n == 0)
    {
      if (p->type == 1)
        printf("+++++++++++++++++++++++++++ CURSOR INFORMATION +++++++++++++++++++++++++++\n\n");
      else if (p->type == 2)
        printf("+++++++++++++++++++++++++++ BITMAP INFORMATION +++++++++++++++++++++++++++\n\n");
      else // (p->type == 3)
        printf("++++++++++++++++++++++++++++ ICON INFORMATION ++++++++++++++++++++++++++++\n\n");
    }
    BITMAPINFOHEADER bh;
    uint32_t xy;
    int len;
    if (p->type == 1)           // read the cursor hotspot
      RawRead(file,xy);
    RawRead(file,bh);
    printf("%4d: ",++p->n);
    if (info->name)
    {
      printf("Name: ");
      len = out_ws(info->name);
    }
    else
    {
      printf("  ID: %04X",info->id);
      len = 4;
    }
    if (len < 26)
      printf("%*s",26-len,"");
    if (p->type != 2)
      bh.Height /= 2;
    printf("%4u x %4u x %2u",bh.Width,bh.Height,bh.BitCount);
    if (p->type == 1)
      printf("   (%u,%u)",xy & 0xFFFF,xy >> 16);
    printf("\n");
  }
}

void ResourceExtract::WriteCursorInfo(PeFile& peFile)
{
  WriteCurBMPIconParam param(1);
  peFile.ParseResource(WriteCurBMPIconInfo,&param);
  if (param.n)
    printf("\n");
}

void ResourceExtract::WriteBitmapInfo(PeFile& peFile)
{
  WriteCurBMPIconParam param(2);
  peFile.ParseResource(WriteCurBMPIconInfo,&param);
  if (param.n)
    printf("\n");
}

void ResourceExtract::WriteIconInfo(PeFile& peFile)
{
  WriteCurBMPIconParam param(3);
  peFile.ParseResource(WriteCurBMPIconInfo,&param);
  if (param.n)
    printf("\n");
}


static void WriteMenuInfo(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  if (data != NULL && level == 3 && info->highLevel->id == 4)
  {
    int* p = reinterpret_cast<int*>(param);
    if (*p == 0)
      printf("++++++++++++++++++++++++++++ MENU INFORMATION ++++++++++++++++++++++++++++\n");
    printf("\n");
    int len;
    len = printf("%4d: ",++*p);
    if (info->name)
    {
      len += printf("Name: ");
      len += out_ws(info->name);
    }
    else
      len += printf("  ID: %04X",info->id);
    printf("\n");
    do
      printf("-");
    while (--len != 0);
    printf("\n");
    uint16_t v,m,i;
    long end = file.Tell() + data->data.Size;
    RawRead(file,v);            // versionNumber
    RawRead(file,i);            // offset
    if (i)
      file.Seek(i,SEEK_CUR);
    level = 0;
    uint32_t id, t;
    int nested = 0;
    if (v == 0 || v == 1)       // MENUITEMTEMPLATE, MENUEX_TEMPLATE_ITEM
    {
      const int popup = (v == 0) ? 0x10 : 0x01;
      do
      {
        if (v == 0)
        {
          RawRead(file,m);      // mtOption
          if (m & 0x10)        // MF_POPUP doesn't have mtID
            i = 0;
          else
            RawRead(file,i);    // mtID
          id = (i == 0xFFFFu) ? 0xFFFFFFFFlu : i;
          t  = m;
        }
        else
        {
          RawRead(file,t);      // dwType
          RawRead(file,id);    // skip dwState
          RawRead(file,id);    // menuId
          RawRead(file,m);      // bResInfo
        }
        printf("%*s",level*4,"");
        unicode_t* pName = reinterpret_cast<unicode_t*>(file.ReadStringFlag(ICFile::strUcs2));
        if (*pName)             // mtString
          len = out_ws(pName);
        else if (len == 0)
          len = printf("<Popup>");
        else if (id && !(t & 0x800)) // MFT_SEPARATOR
          len = printf("<no name>");
        else                    // separator
        {
          for (unsigned n = len; n-- != 0;)
            printf("-");
        }
        if (id != 0u && id != 0xFFFFFFFFlu)
          printf("%*s\t[%04X]",(len+level*4 < 40) ? 40-(len+level*4) : 0,"",(unsigned int) id);
        printf("\n");
        delete [] pName;
        if (m & popup)
        {
          ++level;
          if (v == 1)
            RawRead(file,id);   // skip dwHelpID
          if ((m & 0x80) && level != 1)
            nested = 1;
        }
        else if (m & 0x80)      // MF_END
        {
          level -= 1 + nested;
          if ((int)level <= 0)
            printf("\n");
          nested = 0;
        }
        if (v == 1)
          if (file.Tell() & 2)  // DWORD alignment
            RawRead(file,m);
      } while (file.Tell() < end);
    }
    else
    {
      printf("Unknown menu version: %04X\n\n",v);
    }
  }
}

void ResourceExtract::WriteMenuInfo(PeFile& peFile)
{
  int n = 0;
  peFile.ParseResource(::WriteMenuInfo,&n);
}


static void WriteDialogInfo(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  if (data != NULL && level == 3 && info->highLevel->id == 5)
  {
    int* p = reinterpret_cast<int*>(param);
    if (*p == 0)
      printf("+++++++++++++++++++++++++++ DIALOG INFORMATION +++++++++++++++++++++++++++\n\n");
    int len;
    len = printf("%4d: ",++*p);
    if (info->name)
    {
      len += printf("Name: ");
      len += out_ws(info->name);
    }
    else
      len += printf("  ID: %04X",info->id);
    uint32_t style;
    bool ex;
    DLGTEMPLATE dlg;
    uint16_t i;
    RawRead(file,style);
    if ((style >> 16) == 0xFFFF)        // signature
    {
      if ((style & 0xFFFF) != 1)        // DlgVer
      {
        printf("\nUnknown dialog version: %04X\n\n",style & 0xFFFF);
        return;
      }
      DLGTEMPLATEEX xdlg;
      RawRead(file,xdlg);
      style   = xdlg.style;
      dlg.cdit = xdlg.cDlgItems;
      ex = true;
    }
    else
    {
      RawRead(file,dlg);
      ex = false;
    }
    // read menu & class
    for (int j = 0; j < 2; ++j)
    {
      RawRead(file,i);
      if (i == 0)          // none
        ;
      else if (i == 0xFFFF)  // ordinal
        RawRead(file,i);
      else                  // name
      {
        do
          RawRead(file,i);
        while (i != 0);
      }
    }
    // read caption
    unicode_t* text = reinterpret_cast<unicode_t*>(file.ReadStringFlag(ICFile::strUcs2));
    // read font data, if present
    if (style & DS_SETFONT)
    {
      RawRead(file,i);      // point size
      if (ex)
      {
        RawRead(file,i);    // weight
        RawRead(file,i);    // italic & charset
      }
      do                    // typeface name
        RawRead(file,i);
      while (i != 0);
    }
    if (file.Tell() & 2)    // align first item on DWORD boundary
      RawRead(file,i);
    len += printf("%*s%3d item%c  \"",(len < 21) ? 21-len : 0,"",dlg.cdit,
                  (dlg.cdit == 1) ? ' ' : 's');
    len += out_ws(text);
    len += printf("\"\n");
    while (--len != 0)
      printf("-");
    printf("\n");
    delete [] text;
    DLGITEMTEMPLATE item;
    for (int c = 1; c <= dlg.cdit; ++c)
    {
      if (ex)
      {
        DLGITEMTEMPLATEEX xitem;
        RawRead(file,xitem);
        item.style = xitem.style;
        item.x     = xitem.x;
        item.y     = xitem.y;
        item.id    = xitem.id;
      }
      else
        RawRead(file,item);
      printf("%4d: %04X  ",c,item.id);
      RawRead(file,i);          // control class
      int type = 0;
      if (i == 0xFFFF)
      {
        RawRead(file,i);
        switch (i)
        {
        case 0x80: len = printf("BUTTON");    type = 'b'; break;
        case 0x81: len = printf("EDIT");      break;
        case 0x82: len = printf("STATIC");    type = 's'; break;
        case 0x83: len = printf("LISTBOX");   break;
        case 0x84: len = printf("SCROLLBAR"); break;
        case 0x85: len = printf("COMBOBOX");  break;
        default:   len = printf("%04X",i);    break;
        }
      }
      else
      {
        file.Seek(-2,SEEK_CUR);
        text = reinterpret_cast<unicode_t*>(file.ReadStringFlag(ICFile::strUcs2));
        len = out_ws(text);
        delete [] text;
      }
      if (len < 11)
        len += printf("%*s%3d,%3d",11-len,"",item.x,item.y);
      printf("%*s\t",(len < 12) ? 12-len : 0,"");
      RawRead(file,i);          // control text or ordinal
      if (i == 0xFFFF)
      {
        RawRead(file,i);
        printf("%04X",i);
        if (type == 'b')
        {
          if      (item.style & 0x40) // BS_ICON
            printf(" (icon)");
          else if (item.style & 0x80) // BS_BITMAP
            printf(" (bitmap)");
        }
        else if (type == 's')
        {
          if      ((item.style & 0x1F) == 0x03) // SS_ICON
            printf(" (icon)");
          else if ((item.style & 0x1F) == 0x0E) // SS_BITMAP
            printf(" (bitmap)");
        }
        printf("\n");
      }
      else
      {
        printf("\"");
        file.Seek(-2,SEEK_CUR);
        text = reinterpret_cast<unicode_t*>(file.ReadStringFlag(ICFile::strUcs2));
        out_ws(text);
        printf("\"\n");
        delete [] text;
      }
      RawRead(file,i);          // creation data
      if (i != 0)
      {
        char c;
        for (i -= 2; i; --i)    // does EX include size?
          RawRead(file,c);
        if (file.Tell() & 1)    // WORD alignment
          RawRead(file,c);
      }
      if (file.Tell() & 2)      // DWORD alignment
        RawRead(file,i);
    }
    printf("\n");
  }
}

void ResourceExtract::WriteDialogInfo(PeFile& peFile)
{
  int n = 0;
  peFile.ParseResource(::WriteDialogInfo,&n);
}


static void WriteStringTableInfo(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  if (data != NULL && level == 3 && info->highLevel->id == 6)
  {
    int* p = reinterpret_cast<int*>(param);
    if (*p == 0)
      printf("+++++++++++++++++++++++++++ STRING INFORMATION +++++++++++++++++++++++++++\n\n");
    else if (*p < 0)
      *p = 0;
    for (unsigned n = 0; n < 16; ++n)
    {
      unicode_t* pName = reinterpret_cast<unicode_t*>(file.ReadStringFlag(ICFile::strUcs2|ICFile::strLen16));
      if (*pName)
      {
        printf("%4d: %04X: \"",++*p,((info->id-1) << 4) | n);
        out_ws(pName);
        printf("\"\n");
      }
      delete [] pName;
    }
    if (*p == 0)        // why would it write a resource with no strings?
      --*p;
  }
}

void ResourceExtract::WriteStringTableInfo(PeFile& peFile)
{
  int n = 0;
  peFile.ParseResource(::WriteStringTableInfo,&n);
  if (n)
    printf("\n");
}


static void WriteVersionInformation(ICFile& file,const PeResourceDirectoryData* data,const PeResourceDirectoryInfo* info,unsigned level,void* param)
{
  if (data != NULL && level == 3 && info->highLevel->id == 16)
  {
    printf("++++++++++++++++++++++++++ VERSION INFORMATION +++++++++++++++++++++++++++\n\n");
    uint8_t buf[512];
    // skip VS_VERSION_INFO header
    file.RawRead(buf,0x28);
    VS_FIXEDFILEINFO vs;
    RawRead(file,vs);
    if (vs.dwSignature != 0xFEEF04BD)
    {
      printf("Unrecognized signature: %08X\n\n", vs.dwSignature);
      return;
    }
    #define FIELD "%-24s\t"
    printf(FIELD "%u.%u.%u.%u\n","File version",
                 vs.dwFileVersionMS >> 16, vs.dwFileVersionMS & 0xFFFF,
                 vs.dwFileVersionLS >> 16, vs.dwFileVersionLS & 0xFFFF);
    printf(FIELD "%u.%u.%u.%u\n","Product version",
                 vs.dwProductVersionMS >> 16, vs.dwProductVersionMS & 0xFFFF,
                 vs.dwProductVersionLS >> 16, vs.dwProductVersionLS & 0xFFFF);
    printf(FIELD "%08X","File flags",vs.dwFileFlags);
    const char* comma = " (";
    if (vs.dwFileFlags & 0x01) printf("%sdebug",comma),         comma = ", ";
    if (vs.dwFileFlags & 0x02) printf("%sprerelease",comma),    comma = ", ";
    if (vs.dwFileFlags & 0x04) printf("%spatched",comma),       comma = ", ";
    if (vs.dwFileFlags & 0x08) printf("%sprivate build",comma), comma = ", ";
    if (vs.dwFileFlags & 0x10) printf("%sinfo inferred",comma), comma = ", ";
    if (vs.dwFileFlags & 0x20) printf("%sspecial build",comma), comma = ", ";
    if (*comma == ',')
      printf(")");
    printf("\n");
    printf(FIELD "%08X (","File OS",vs.dwFileOS);
    if (vs.dwFileOS == 0)
      printf("unknown");
    else
    {
      switch(vs.dwFileOS >> 16)
      {
      case 1: printf("DOS");         break;
      case 2: printf("OS/2 16-bit"); break;
      case 3: printf("OS/2 32-bit"); break;
      case 4: printf("Windows NT");  break;
      case 5: printf("Windows CE");  break;
      }
      if ((vs.dwFileOS >> 16) && (vs.dwFileOS & 0xFFFF))
        printf(" / ");
      switch(vs.dwFileOS & 0xFFFF)
      {
      case 1: printf("Windows 16-bit");              break;
      case 2: printf("Presentation Manager 16-bit"); break;
      case 3: printf("Presentation Manager 32-bit"); break;
      case 4: printf("Windows 32-bit");              break;
      }
    }
    printf(")\n");
    printf(FIELD "%08X (","File type",vs.dwFileType);
    switch(vs.dwFileType)
    {
    default:
    case 0: printf("unknown");        break;
    case 1: printf("application");    break;
    case 2: printf("DLL");            break;
    case 3: printf("device driver");  break;
    case 4: printf("font");           break;
    case 5: printf("virtual device"); break;
    case 7: printf("static library"); break;
    }
    printf(")\n");
    if (vs.dwFileSubtype)
    {
      printf(FIELD "%8X (","File subtype",vs.dwFileSubtype);
      switch(vs.dwFileType)
      {
      default: printf("unknown"); break;
      case 3:
        switch(vs.dwFileSubtype)
        {
        default:
        case 0x00: printf("unknown");           break;
        case 0x01: printf("printer");           break;
        case 0x02: printf("keyboard");          break;
        case 0x03: printf("language");          break;
        case 0x04: printf("display");           break;
        case 0x05: printf("mouse");             break;
        case 0x06: printf("network");           break;
        case 0x07: printf("system");            break;
        case 0x08: printf("installable");       break;
        case 0x09: printf("sound");             break;
        case 0x0A: printf("communications");    break;
        case 0x0B: printf("inputmethod");       break;
        case 0x0C: printf("versioned printer"); break;
        }
        break;
      case 4:
        switch(vs.dwFileSubtype)
        {
        default:
        case 0: printf("unknown");  break;
        case 1: printf("raster");   break;
        case 2: printf("vector");   break;
        case 3: printf("truetype"); break;
        }
        break;
      }
      printf(")\n");
    }
    if (vs.dwFileDateMS || vs.dwFileDateLS)
      printf(FIELD "%08X%08X\n","File creation date",
                   vs.dwFileDateMS, vs.dwFileDateLS);
    file.RawRead(buf,0x24);
    if (memcmp(buf+6,L"StringFileInfo",0x1E) != 0)
      return;
    uint16_t len;
    RawRead(file,len);
    file.RawRead(buf,0x16);
    len -= 0x18;
    if (len & 2)
      len += 2;
    printf(FIELD,"Language");
    out_ws(reinterpret_cast<const unicode_t*>(buf+4));
    printf("\n");
    while (len)
    {
      uint16_t sz;
      RawRead(file,sz);
      if (sz & 2)
        sz += 2;
      _PRG_ASSERT(sz-2 < 512);
      file.RawRead(buf,sz-2);
      uint32_t n = out_ws(reinterpret_cast<const unicode_t*>(buf+4));
      printf("%*s\t",((n & 0xFFFF) < 24) ? 24-(n & 0xFFFF) : 0,"");
      n >>= 16;
      if (n & 1)
        ++n;
      if (*reinterpret_cast<uint16_t*>(buf))
        out_ws(reinterpret_cast<const unicode_t*>(buf+4+(n+1)*2));
      printf("\n");
      len -= sz;
    }
    printf("\n");
  }
  #undef FIELD
}

void ResourceExtract::WriteVersionInformationInfo(PeFile& peFile)
{
  peFile.ParseResource(::WriteVersionInformation,NULL);
}


void ResourceExtract::WriteAnimatedCursorInfo(PeFile& peFile)
{
  // wait till I find any
}


void ResourceExtract::WriteAnimatedIconInfo(PeFile& peFile)
{
  // wait till I find any
}
