/*
PeRdr - PE file disassembler
Copyright (C) 1999-2002 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#ifndef FILE_REFERENCES_H
#define FILE_REFERENCES_H

#include <set>

enum FlowTypes { FLOW_NONE, FLOW_RET, FLOW_CALL, FLOW_JUMP, FLOW_CJUMP, FLOW_CONST };

class IstrReferences;

class IstrReference
{
public:
  IstrReference() {};

  vma_t from;
  vma_t to;
  enum FlowTypes type;
  bool direct;
  bool isApi;
  bool operator==(const IstrReference& rhs) const
  { return from==rhs.from && to==rhs.to; };
  bool operator<(const IstrReference& rhs) const
  { return from<rhs.from || (from==rhs.from && to<rhs.to); }

private:

  friend class IstrReferences;

  IstrReference(vma_t _from,vma_t _to,bool _direct,enum FlowTypes _type,bool _isApi=false):
    from(_from),to(_to),type(_type),direct(_direct),isApi(_isApi) {};
  IstrReference(vma_t _from,vma_t _to):
    from(_from),to(_to) {};
};

class IstrReferences
{
  typedef std::set< IstrReference,std::less<IstrReference> > TRefs;
  TRefs refs;
public:
  typedef TRefs::const_iterator iterator;
  void Add(vma_t from, vma_t to, enum FlowTypes type, bool isApi = false);
  bool IsPresent(vma_t from,vma_t to) const;
  iterator beginReferenceTo(vma_t addr) const;
  iterator endReferenceTo(vma_t addr) const;
};

#endif // FILE_REFERENCES_H
