/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include "peformat.h"
#include "rva.h"

void PeSection::operator=(const PE_IMAGE_SECTION_HEADER& section)
{
  RvaSize         = section.Misc.VirtualSize;
  RvaAddress      = section.VirtualAddress;
  RawSize         = section.SizeOfRawData;
  RawAddress      = section.PointerToRawData;
  Characteristics = section.Characteristics;
}

bool PeSection::IsCode() const
{
  return (Characteristics & IMAGE_SCN_CNT_CODE) != 0;
}

RVAFileTranslator::RVAFileTranslator(const PE_IMAGE_SECTION_HEADER* _sections,unsigned _num_section):
  num(_num_section),section_infos(NULL)
{
  unsigned n;

  // check order and size
  uint32_t prevVirtualEnd = 0,prevFileEnd = 0;
  for( n=0; n<num; ++n )
  {
    // vedi se sezione inizializzata
    if (!(_sections[n].Characteristics&IMAGE_SCN_CNT_UNINITIALIZED_DATA) || _sections[n].PointerToRawData != 0)
      // !!! una sezione puo' essere vuota e non avere il flag uninitialized settato
      if(_sections[n].PointerToRawData != 0 || _sections[n].SizeOfRawData != 0)
      {
        if (prevVirtualEnd > _sections[n].VirtualAddress ||
            prevFileEnd > _sections[n].PointerToRawData )
          throw std::runtime_error("Invalid order of sections");
        prevVirtualEnd = _sections[n].VirtualAddress+_sections[n].Misc.VirtualSize;
        prevFileEnd    = _sections[n].PointerToRawData+_sections[n].SizeOfRawData;
      }
  }

  // copy
  section_infos = new PeSection[num];
  for( n=0; n<num; ++n )
    section_infos[n] = _sections[n];
}

RVAFileTranslator::~RVAFileTranslator()
{
  delete[] section_infos;
}

long RVAFileTranslator::RVA2File(uint32_t rva) const
{
  for( unsigned n=0; n<num; ++n )
  {
    if (rva >= section_infos[n].RvaAddress && rva < (section_infos[n].RvaAddress+section_infos[n].RvaSize) )
      return rva - section_infos[n].RvaAddress + section_infos[n].RawAddress;
  }
  return 0;
}

uint32_t RVAFileTranslator::File2RVA(long _file) const
{
  if (_file < 0) return 0;
  unsigned long file = static_cast<unsigned long>(_file);
  for( unsigned n=0; n<num; ++n )
  {
    if (file >= section_infos[n].RawAddress && file < (section_infos[n].RawAddress+section_infos[n].RawSize) )
      return file - section_infos[n].RawAddress + section_infos[n].RvaAddress;
  }
  return 0;
}

void RVAFileTranslator::operator=(const RVAFileTranslator& rhs)
{
  delete section_infos;
  section_infos = new PeSection[rhs.num];
  num = rhs.num;
  for(unsigned n=0; n<num; ++n)
    section_infos[n] = rhs.section_infos[n];
}

