/*
 * PeRdr - PE file disassembler
 * Copyright (C) 1999-2006 Frediano Ziglio
 * -----
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * -----
 *
 * INFORMATION
 *   www: http://freddy77.tripod.com/perdr/
 *     e-mail: freddy77@angelfire.com
 *     */
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include "utils.hpp"

#ifdef WIN32

#undef ULONG
#undef DWORD
#undef INT
#undef UINT
#undef LONG
#include <windows.h>

std::string FindDll(const char *dllName)
{
  char filename[MAX_PATH];

  if (!SearchPath(NULL, dllName, ".dll", sizeof(filename), filename, NULL) 
    && !SearchPath(NULL, dllName, ".ocx", sizeof(filename), filename, NULL))
    return "";
  return filename;
}
#else

#include <dirent.h>

std::string FindDll(const char *dllName)
{
  // !!! for now search in current directory
  DIR *dir = opendir(".");
  if (!dir)
    return "";
  
  struct dirent* file;
  int len = strlen(dllName);
  while( (file=readdir(dir)) != NULL )
  {
    if (strncasecmp(file->d_name, dllName, len) != 0)
      continue;
    if (file->d_name[len] != '.')
      continue;
    if (strcasecmp(file->d_name+len+1,"dll") == 0 || strcasecmp(file->d_name+len+1,"ocx") == 0)
    {
      closedir(dir);
      return file->d_name;
    }
  }
  closedir(dir);
  return "";
}
#endif
