/*
fix size vector class similar to STL
Copyright (C) 1999-2002 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#ifndef FILE_FIXVECT_HPP
#define FILE_FIXVECT_HPP

template <class T,unsigned _size>
class fix_vector
{
public:
  typedef T         value_type;
  typedef T&        reference;
  typedef const T&  const_reference;
  typedef T*        iterator;
  typedef const T*  const_iterator;
  typedef size_t    size_type;
  typedef ptrdiff_t difference_type;
#if defined(__GNUC__) || (defined(_MSC_VER) && _MSC_VER >= 1300)
  typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
  typedef std::reverse_iterator<iterator>       reverse_iterator;
#else
  typedef std::reverse_iterator<const_iterator, value_type,
                             const_reference, difference_type>
                                                       const_reverse_iterator;
  typedef std::reverse_iterator<iterator, value_type,
                             reference, difference_type>
                                                       reverse_iterator;
#endif

  // iterator access
  iterator       begin()       { return info; }
  const_iterator begin() const { return info; }
  iterator       end  ()       { return info+_size; }
  const_iterator end  () const { return info+_size; }

  // reverse iterator access
  reverse_iterator rbegin ()
  {
    return reverse_iterator(end());
  }
  const_reverse_iterator rbegin () const
  {
    return const_reverse_iterator(end());
  }
  reverse_iterator rend ()
  {
    return reverse_iterator(begin());
  }
  const_reverse_iterator rend () const
  {
    return const_reverse_iterator(begin());
  }

  // capacity
  size_type size() const { return _size; }
  size_type capacity () const { return _size; }
  bool empty () const { return false; }

  // element access
  reference       operator[] (size_type n)       { return info[n]; }
  const_reference operator[] (size_type n) const { return info[n]; }
  reference       front ()                       { return info[0]; }
  const_reference front ()                 const { return info[0]; }
  reference       back ()                        { return *(end() - 1); }
  const_reference back ()                  const { return *(end() - 1); }
private:
  T info[_size];
};

#endif //FILE_FIXVECT_HPP

